FLIGHTMAP_MAX_TAXIPATHS = 48;
FLIGHTMAP_MAX_TAXINODES = 32;

function FlightMapTaxi_SetContinent(cont)
    local cname = FlightMapUtil.getContinentName(cont);
    TaxiMerchant:SetText(cname);
    local tname = "Interface\\TaxiFrame\\TAXIMAP" .. 2-cont;
    TaxiMap:SetTexture(tname);
    FlightMapTaxiFrame_OnEvent(cont);
end

function FlightMapTaxi_ShowContinent(cont)
    if TaxiFrame:IsVisible() then return; end
    TaxiPortrait:SetTexture("Interface\\WorldMap\\WorldMap-Icon");
    for i = 1, TAXI_BUTTONS, 1 do
        local btn = getglobal("TaxiButton" .. i);
        btn:Hide();
    end
    FlightMapTaxi_SetContinent(cont);
    TaxiRouteMap:Hide();
    FlightMapTaxiContinents:Show();
    ShowUIPanel(TaxiFrame, 1);
    UIDropDownMenu_SetSelectedID(FlightMapTaxiContinents, cont);
end

function FlightMapTaxiFrame_OnUpdate()
    -- Do one iteration over the to-visit list
    local node = table.remove(this.to_visit, this.to_visit.n);
    local map = FlightMapUtil.getFlightMap();

    if node then
        -- Note that this node has been visited
        this.visited[node.key] = true;

        -- Node will be {key: ..., time: ..., history: {...}}
        if map[node.key] then
            for k, v in map[node.key].Flights do
                -- Only to known nodes, only for known times, not for
                -- already visited nodes
                if not this.visited[k] and FlightMapUtil.knownNode(k)
                and v and v > 0 then
                    -- Create a new step in the depth first graph walk
                    local step = {
                        key = k,
                        time = node.time + v,
                        history = {n = 0},
                    };
                    for i = 1, node.history.n, 1 do
                        table.insert(step.history, node.history[i]);
                    end
                    table.insert(step.history, {dest = k, time = v});

                    -- See if there is an existing copy of this destination
                    local old = nil;
                    for i = 1, this.to_visit.n, 1 do
                        if this.to_visit[i].key == k then
                            old = i;
                        end
                    end

                    -- If any existing flight is longer, remove it
                    if old and this.to_visit[old].time > step.time then
                        table.remove(this.to_visit, old);
                        old = nil;
                    end

                    -- If there's no existing flight, gotta add one
                    if not old then
                        -- Insert step into to-visit table, sorted by time
                        local spot = this.to_visit.n + 1;
                        while spot > 1
                        and step.time > this.to_visit[spot - 1].time do
                            spot = spot - 1;
                        end
                        table.insert(this.to_visit, spot, step);

                        -- Store the quickest route to this node
                        if FlightMapTaxiFrame.Buttons[k] then
                            FlightMapTaxiFrame.Buttons[k].quickest =
                                    step.history;
                        end
                    end
                end
            end
        end
    else
        -- Remove OnUpdate handler
        this:SetScript("OnUpdate", nil);
    end
end

function FlightMapTaxiButton_OnEnter(button)
    GameTooltip:SetOwner(button, "ANCHOR_RIGHT");

    -- Add in this node's flight details
    FlightMapUtil.addFlightsForNode(GameTooltip, button.nodeKey, '');

    -- Check for quickest route data
    if button.quickest then
        local map = FlightMapUtil.getFlightMap();
        local total = 0;

        GameTooltip:AddLine("\n");
        GameTooltip:AddLine(FLIGHTMAP_QUICKEST, 1.0, 1.0, 1.0);
        for i = 1, button.quickest.n, 1 do
            local data = button.quickest[i];
            GameTooltip:AddDoubleLine("  " .. map[data.dest].Name,
                FlightMapUtil.formatTime(data.time),
                0.7, 0.7, 0.7, 0.7, 0.7, 0.7);
            total = total + data.time
        end
        GameTooltip:AddDoubleLine("  " .. FLIGHTMAP_TOTAL_TIME,
            FlightMapUtil.formatTime(total),
            1.0, 1.0, 1.0, 1.0, 1.0, 1.0);
    end

    -- "fade in" the flights connected to this node
    for k, v in button.lines do
        v:SetAlpha(1.0);
    end

    GameTooltip:Show();
end

function FlightMapTaxiButton_OnLeave(button)
    GameTooltip:Hide();

    -- "fade out" the flights connected to this node
    for k, v in button.lines do
        v:SetAlpha(0.5);
    end
end

function FlightMapTaxiButton_OnClick(button)
    -- Take a flight to the first destination in the quickest path
    if button.quickest then
        local id = FlightMapTaxiFrame.Destinations[button.quickest[1].dest];
        if id then
            TakeTaxiNode(id);
        end
    end
end

function FlightMapTaxiFrame_OnLoad()
    this:RegisterEvent("TAXIMAP_OPENED");
    TaxiButtonTypes["REMOTE"] = {
        file = "Interface\\AddOns\\FlightMap\\UI-Taxi-Icon-Blue";
    }
    TaxiButtonTypes["UNKNOWN"] = {
        file = "Interface\\TaxiFrame\\UI-Taxi-Icon-Gray";
    }
end

function FlightMapTaxiFrame_OnEvent(event)
    -- Hide any unused lines left over from a previous flight master visit
    for i = 1, FLIGHTMAP_MAX_TAXIPATHS, 1 do
        getglobal("FlightMapTaxiPath" .. i):Hide();
    end

    -- Hide any unused icons too
    for i = 1, FLIGHTMAP_MAX_TAXINODES, 1 do
        getglobal("FlightMapTaxiButton" .. i):Hide();
    end

    -- If the event isn't "TAXIMAP_OPENED" then this is an
    -- anywhere-display map.
    local thiscont = event;
    if event == "TAXIMAP_OPENED" then
        -- Turn the route map back on
        TaxiRouteMap:Show();
        -- Turn the continent select off
        FlightMapTaxiContinents:Hide();
        -- Check option is on
        if not FlightMap.Opts.fullTaxiMap then return; end
        thiscont = FlightMapUtil.getContinent();
    end

    local map = FlightMapUtil.getFlightMap();

    -- Build a list of already displayed nodes
    local shownNodes = {};
    local currentNode;
    FlightMapTaxiFrame.Destinations = {};
    for i = 1, NumTaxiNodes() do
        local tx, ty = TaxiNodePosition(i);
        local node = FlightMapUtil.makeNodeName(thiscont, tx, ty);
        shownNodes[node] = true;
        if TaxiNodeGetType(i) == "CURRENT" then
            currentNode = node;
        end
        FlightMapTaxiFrame.Destinations[node] = i;
    end

    -- Keep track of which flights have been drawn in already
    local linesdrawn = {};
    local linenum = 1;

    -- Keep track of which button is for which taxi node
    FlightMapTaxiFrame.Buttons = {};

    local nodenum = 1;
    for key, node in map do
        -- If it's not already shown, and is on this continent ...
        if not shownNodes[key] and node.Continent == thiscont then
            -- Make sure there's room to show it
            if nodenum < FLIGHTMAP_MAX_TAXINODES then
                local button = getglobal("FlightMapTaxiButton" ..  nodenum);
                button:ClearAllPoints();
                button:SetPoint("CENTER", "TaxiMap", "BOTTOMLEFT",
                    node.Location.Taxi.x * TAXI_MAP_WIDTH,
                    node.Location.Taxi.y * TAXI_MAP_HEIGHT);
                local nType = "NONE";
                if FlightMapUtil.knownNode(key) then
                    nType = "REMOTE";
                elseif FlightMap.Opts.showAllInfo then
                    nType = "UNKNOWN";
                end
                if nType ~= "NONE" then
                    -- Store data for tooltips
                    button.nodeKey  = key;
                    button.node     = node;
                    button.quickest = nil;
                    button.cheapest = nil;
                    button.lines    = {};

                    -- Store against the taxi frame for later access
                    FlightMapTaxiFrame.Buttons[key] = button;

                    -- Oh, and display it, I guess
                    button:SetNormalTexture(TaxiButtonTypes[nType].file);
                    button:Show();

                    nodenum = nodenum + 1;

                    -- For each destination
                    for k, v in node.Flights do
                        local line = key .. "-" .. k;
                        local back = k .. "-" .. key;
                        if not linesdrawn[line]
                        and not linesdrawn[back]
                        and linenum < FLIGHTMAP_MAX_TAXIPATHS
                        and (FlightMapUtil.knownNode(k) 
                         or FlightMap.Opts.showAllInfo) then
                            local tex = getglobal("FlightMapTaxiPath"
                                    .. linenum);
                            local dest = map[k];
                            if dest then
                                linesdrawn[line] = tex;
                                FlightMapUtil.drawLine(TaxiMap, tex,
                                    node.Location.Taxi.x,
                                    1 - node.Location.Taxi.y,
                                    dest.Location.Taxi.x,
                                    1 - dest.Location.Taxi.y);
                                tex:SetAlpha(0.5);
                                linenum = linenum + 1;
                            end
                        end
                        -- Store the line against the button's list
                        if not linesdrawn[line] then line = back; end
                        if linesdrawn[line] then
                            button.lines[line] = linesdrawn[line];
                        end
                    end
                end
            end
        end
    end

    -- Start an OnUpdate handler to fill in the quickest route
    -- information without lagging the user up
    if currentNode ~= nil then
        FlightMapTaxiFrame.to_visit = {
            {key = currentNode, time = 0, history = {n = 0}}, n = 1
        };
        this.visited = {};
        this:SetScript("OnUpdate", FlightMapTaxiFrame_OnUpdate);
    end
end

function FlightMapTaxiContinents_OnLoad()
    UIDropDownMenu_Initialize(this, FlightMapTaxiContinents_Initialize);
    UIDropDownMenu_SetWidth(130);
end

function FlightMapTaxiContinents_Initialize()
    local function init(...)
        local info;
        for i = 1, arg.n, 1 do
            info = {
                text = arg[i],
                func = FlightMapTaxiContinents_OnClick,
            };
            UIDropDownMenu_AddButton(info);
        end
    end
    init(GetMapContinents());
end

function FlightMapTaxiContinents_OnClick()
    UIDropDownMenu_SetSelectedID(FlightMapTaxiContinents, this:GetID());
    FlightMapTaxi_SetContinent(this:GetID());
end
